import speech_recognition as sr
from .sapi import Sapi
from io import BytesIO
import time
import os
import pyaudio
import keyboard
import threading
from time import sleep
import numbers

class VoiceRobotifier:

    def __init__(self):
        self.microphone = sr.Microphone()
        self.recognizer = sr.Recognizer()
        self.voice = Sapi()

    def set_input_device(self,device_name):
        device_id = None
        for i,name in enumerate(self.list_device_names()):
            if name == device_name[:31]:
                device_id = i
                break
        if not device_id:
            raise Exception('Device name "%s" not found' % device_input)
        self.microphone = sr.Microphone(device_id)
            

    def set_output_device(self,device_name):
        self.voice.set_audio_output(device_name)

    def list_device_names(self):
        audio = pyaudio.PyAudio()
        out = list()
        for i in range(0,audio.get_device_count()):
            out.append(audio.get_device_info_by_index(i)['name'])
        return out

    def set_voice(self,voice_name):
        self.voice.set_voice(voice_name)

    def list_voices(self):
        return self.voice.get_voice_names()

    def set_voice_rate(self,voice_rate):
        self.voice.set_rate(voice_rate)
        

    def start(self,key,key_quit='esc',block=True):
        running = Event()
        lock = Event()

        def stop():
            nonlocal lock, running
            lock.stop()
            running.set()
            keyboard.unhook(key_quit)
            print('Stopped VoiceRobotifier')

        def target():
            while not running.is_set():
                lock.clear()
                keyboard.on_press_key(key, lambda e: lock.set())

                lock.wait()
                keyboard.unhook(key)

                lock.clear()
                if running.is_set():
                    break
                keyboard.on_release_key(key, lambda e: lock.set())
                
                print('Listening...', end=' ')
                frames = BytesIO()
                with self.microphone as source:
                    while not lock.is_set():  # loop for the total number of chunks needed
                        if source.stream is None or source.stream.pyaudio_stream.is_stopped(): break
                        buffer = source.stream.read(source.CHUNK)
                        if len(buffer) == 0: break
                        frames.write(buffer)
                    keyboard.unhook(key)
                    
                    frame_data = frames.getvalue()
                    frames.close()
                    audio = sr.AudioData(frame_data, source.SAMPLE_RATE, source.SAMPLE_WIDTH)
                self.__audio_to_tts(audio)

        if key_quit is not None:
            keyboard.on_press_key(key_quit,lambda e: stop())

        thread = threading.Thread(target=target,daemon=True)
        thread.start()
        print('Started VoiceRobotifier with push-to-talk key \''+str(key)+'\'')

        if block:
            thread.join()
        
        return stop


    def start_autodetect(self,energy_threshold=None,key_quit='esc',block=True):
        blocking = Event()
        
        if energy_threshold is None:
            with self.microphone as source:
                self.recognizer.adjust_for_ambient_noise(source)
        else:
            self.recognizer.energy_threshold = energy_threshold

        def callback(recognizer,audio):
            self.__audio_to_tts(audio)
            print('Listening...', end=' ')

        print('Started VoiceRobotifier with energy threshold '+str(self.recognizer.energy_threshold))
        print('Listening...', end=' ')
        stop_listener = self.recognizer.listen_in_background(self.microphone, callback)

        def stop():
            nonlocal blocking
            stop_listener()
            blocking.set()
            keyboard.unhook(key_quit)
            print('Stopped VoiceRobotifier')
        
        if key_quit is not None:
            keyboard.on_press_key(key_quit,lambda e: stop())
        
        if block:
            blocking.wait()
        
        return stop


    def __audio_to_tts(self,audio):
        try:
            text = uncensor(self.recognizer.recognize_google(audio))
            print('"%s"' % text)
            speak_thread = threading.Thread(target=lambda: self.voice.say(text))
            speak_thread.start()
        except sr.UnknownValueError:
            print("Google Speech Recognition could not understand audio")
        except sr.RequestError as e:
            print("Could not request results from Google Speech Recognition service; {0}".format(e))


class Event(threading.Event):
    def __init__(self):
        self.__stopped = False
        super().__init__()
    def wait(self):
        while True:
            if super().wait(0.5) or self.__stopped:
                break
    def is_set(self):
        if self.__stopped:
            return True
        return super().is_set()
    isSet = is_set
    def stop(self):
        self.__stopped = True
         
def uncensor(text):
    censored = {
        'f***': 'fuck',
        's***': 'shit',
        'b****': 'bitch',
        'a**': 'ass'
    }
    for key,val in censored.items():
        text = text.replace(key,val)
    return text




import os # Allows checking if using Windows
try:
    assert(os.name == 'nt') # Checks for Windows
except:
    raise RuntimeError("Windows is required.")

import comtypes.client  # Importing comtypes.client will make the gen subpackage
try:
    from comtypes.gen import SpeechLib  # comtypes
except ImportError:
    # Generate the SpeechLib lib and any associated files
    engine = comtypes.client.CreateObject("SAPI.SpVoice")
    stream = comtypes.client.CreateObject("SAPI.SpFileStream")
    from comtypes.gen import SpeechLib


class Sapi(object):
    """A speech API using the Microsoft SAPI through COM"""

    def __init__(self):
        super().__init__()
        self.voice = comtypes.client.CreateObject('Sapi.SpVoice')
        self.voice.AudioOutput.GetDescription()

    def get_voices(self, name=''):
        """Get a list of voices, search by name optional"""
        voice_list = []
        voices = self.voice.GetVoices()

        if name is not '':
            for voice in voices:
                if name in voice.GetDescription():
                    voice_list.append(voice)
                    break
            else:
                print('Voice not found')
        else:
            for voice in voices:
                voice_list.append(voice)

        return voice_list

    def get_voice_names(self):
        """Get the names of all the voices"""
        return [voice.GetDescription() for voice in self.get_voices()]

    def set_voice(self, voice):
        """Set the voice to the given voice"""
        if type(voice) is str:
            self.voice.Voice = self.get_voices(voice)[0]
        else:
            self.voice.Voice = voice
        return

    def get_audio_outputs(self, name=''):
        """Get the audio outputs, search for the one with the name if given"""
        output_list = []
        outputs = self.voice.GetAudioOutputs()

        if name is not '':
            for output in outputs:
                if name in output.GetDescription():
                    output_list.append(output)
                    break
            else:
                print('Audio output not found')
        else:
            for output in outputs:
                output_list.append(output)

        return output_list

    def get_audio_output_names(self):
        """Get the names of all the audio outpus"""
        return [output.GetDescription() for output in self.get_audio_outputs()]

    def set_audio_output(self, output):
        if type(output) is str:
            self.voice.AudioOutput = self.get_audio_outputs(output)[0]
        else:
            self.voice.AudioOutput = output
        return

    def say(self, message):
        self.voice.Speak(message)
        return
    
    def set_rate(self, rate):
        """Set the speed of the speaker
        -10 is slowest, 10 is fastest"""
        self.voice.Rate = rate

    def _create_stream(self, filename):
        """Create a file stream handler"""
        stream = comtypes.client.CreateObject('Sapi.SpFileStream')
        stream.Open(filename, SpeechLib.SSFMCreateForWrite)
        return stream

    def create_recording(self, filename, message):
        """Make a recording of the given message to the file
        The file should be a .wav as the output is
        PCM 22050 Hz 16 bit, Little endianness, Signed"""
        stream = self._create_stream(filename)
        temp_stream = self.voice.AudioOutputStream
        self.voice.AudioOutputStream = stream
        self.say(message)
        self.voice.AudioOutputStream = temp_stream


if __name__ == '__main__':
    v = Sapi()
    v.set_voice('Anna')
    v.get_voice_names()



from setuptools import setup

setup(name='voice_robotifier',
        version='1.0',
        description='Turns your voice into text-to-speech!',
        author='Reid Huntley',
        author_email='reidbhuntley@gmail.com',
        url='https://github.com/reidbhuntley/voice_robotifier',
        license='MIT',
        packages=['voice_robotifier'],
        install_requires=[
            'SpeechRecognition>=3.8.1',
            'keyboard>=0.13.2',
            'PyAudio>=0.2.11',
            'comtypes'
        ],
        zip_safe=False)
